"""Generator for Makefile and CMakeLists.txt"""
import pathlib
from typing import List, Optional, Union

from .contrib import gcc, msvc
from .contrib.util import _libext, _toolchain_exist_check
from .exception import TL2cgenError
from .util import _open_and_validate_recipe, _process_options


def generate_makefile(
    dirpath: Union[str, pathlib.Path],
    toolchain: str,
    options: Optional[List[str]] = None,
) -> None:
    """
    Generate a Makefile for a given directory of headers and sources. The
    resulting Makefile will be stored in the directory. This function is useful
    for deploying a model on a different machine.

    Parameters
    ----------
    dirpath :
        Directory containing the header and source files previously generated
        by :py:meth:`Model.compile`. The directory must contain recipe.json
        which specifies build dependencies.
    toolchain :
        Which toolchain to use. You may choose one of 'msvc', 'clang', and 'gcc'.
        You may also specify a specific variation of clang or gcc (e.g. 'gcc-7')
    options :
        Additional options to pass to toolchain
    """
    # pylint: disable=R0912,R0914,W0212
    dirpath = pathlib.Path(dirpath).expanduser().resolve()
    if not dirpath.is_dir():
        raise TL2cgenError(f"Directory {dirpath} does not exist")
    recipe = _open_and_validate_recipe(dirpath / "recipe.json")
    options = _process_options(options)

    # Determine file extensions for object and library files
    _toolchain_exist_check(toolchain)
    if toolchain == "msvc":
        _obj_ext = msvc._obj_ext
        _obj_cmd = msvc._obj_cmd
        _lib_cmd = msvc._lib_cmd
    else:
        _obj_ext = gcc._obj_ext
        _obj_cmd = gcc._obj_cmd
        _lib_cmd = gcc._lib_cmd
    obj_ext = _obj_ext()
    lib_ext = _libext()

    with open(dirpath / "Makefile", "w", encoding="UTF-8") as f:
        objects = [x["name"] + obj_ext for x in recipe["sources"]]
        objects.extend(recipe.get("extra", []))
        target = recipe["target"] + lib_ext
        objects_str = " ".join(objects)
        lib_cmd = _lib_cmd(
            objects=objects,
            target=recipe["target"],
            lib_ext=lib_ext,
            toolchain=toolchain,
            options=options,
        )

        print(f"{target}: {objects_str}", file=f)
        print(f"\t{lib_cmd}", file=f)
        for source in recipe["sources"]:
            source_file = source["name"] + ".c"
            obj_file = source["name"] + obj_ext
            obj_cmd = _obj_cmd(
                source=source["name"], toolchain=toolchain, options=options
            )
            print(f"{obj_file}: {source_file}", file=f)
            print(f"\t{obj_cmd}", file=f)


def generate_cmakelists(
    dirpath: Union[str, pathlib.Path],
    options: Optional[List[str]] = None,
) -> None:
    """
    Generate a CMakeLists.txt for a given directory of headers and sources. The
    resulting CMakeLists.txt will be stored in the directory. This function is useful
    for deploying a model on a different machine.

    Parameters
    ----------
    dirpath :
        Directory containing the header and source files previously generated
        by :py:meth:`Model.compile`. The directory must contain recipe.json
        which specifies build dependencies.
    options :
        Additional options to pass to toolchain
    """
    dirpath = pathlib.Path(dirpath).expanduser().resolve()
    if not dirpath.is_dir():
        raise TL2cgenError(f"Directory {dirpath} does not exist")
    recipe = _open_and_validate_recipe(dirpath / "recipe.json")
    options = _process_options(options)

    target = recipe["target"]
    sources = " ".join([x["name"] + ".c" for x in recipe["sources"]])
    options_str = " ".join(options)
    with open(dirpath / "CMakeLists.txt", "w", encoding="UTF-8") as f:
        print("cmake_minimum_required(VERSION 3.13)", file=f)
        print("project(mushroom LANGUAGES C)\n", file=f)
        print(f"add_library({target} SHARED)", file=f)
        print(f"target_sources({target} PRIVATE header.h {sources})", file=f)
        print(f"target_compile_options({target} PRIVATE {options_str})", file=f)
        print(
            f'target_include_directories({target} PRIVATE "${{PROJECT_BINARY_DIR}}")',
            file=f,
        )
        print(f"set_target_properties({target} PROPERTIES", file=f)
        print(
            """POSITION_INDEPENDENT_CODE ON
            C_STANDARD 99
            C_STANDARD_REQUIRED ON
            PREFIX ""
            RUNTIME_OUTPUT_DIRECTORY "${PROJECT_BINARY_DIR}"
            RUNTIME_OUTPUT_DIRECTORY_DEBUG "${PROJECT_BINARY_DIR}"
            RUNTIME_OUTPUT_DIRECTORY_RELEASE "${PROJECT_BINARY_DIR}"
            RUNTIME_OUTPUT_DIRECTORY_RELWITHDEBINFO "${PROJECT_BINARY_DIR}"
            RUNTIME_OUTPUT_DIRECTORY_MINSIZEREL "${PROJECT_BINARY_DIR}"
            LIBRARY_OUTPUT_DIRECTORY "${PROJECT_BINARY_DIR}"
            LIBRARY_OUTPUT_DIRECTORY_DEBUG "${PROJECT_BINARY_DIR}"
            LIBRARY_OUTPUT_DIRECTORY_RELEASE "${PROJECT_BINARY_DIR}"
            LIBRARY_OUTPUT_DIRECTORY_RELWITHDEBINFO "${PROJECT_BINARY_DIR}"
            LIBRARY_OUTPUT_DIRECTORY_MINSIZEREL "${PROJECT_BINARY_DIR}")
        """,
            file=f,
        )



"""Launcher for C compiler to build shared libs"""
import pathlib
import time
import warnings
from multiprocessing import cpu_count
from typing import List, Optional, Union

from .contrib.gcc import _create_shared_gcc
from .contrib.msvc import _create_shared_msvc
from .contrib.util import _toolchain_exist_check
from .exception import TL2cgenError
from .util import _open_and_validate_recipe, _process_options


def create_shared(
    toolchain: str,
    dirpath: Union[str, pathlib.Path],
    *,
    nthread: Optional[int] = None,
    verbose: bool = False,
    options: Optional[List[str]] = None,
    long_build_time_warning: bool = True,
):  # pylint: disable=R0914
    """Create shared library.

    Parameters
    ----------
    toolchain :
        Which toolchain to use. You may choose one of "msvc", "clang", and "gcc".
        You may also specify a specific variation of clang or gcc (e.g. "gcc-7")
    dirpath :
        Directory containing the header and source files previously generated
        by :py:meth:`generate_c_code`. The directory must contain recipe.json
        which specifies build dependencies.
    nthread :
        Number of threads to use in creating the shared library.
        Defaults to the number of cores in the system.
    verbose :
        Whether to produce extra messages
    options :
        Additional options to pass to toolchain
    long_build_time_warning :
        If set to False, suppress the warning about potentially long build time

    Returns
    -------
    libpath :
        Absolute path of created shared library

    Example
    -------
    The following command uses Visual C++ toolchain to generate
    ``./my/model/model.dll``:

    .. code-block:: python

        tl2cgen.generate_c_code(model, dirpath="./my/model",
                                params={})
        tl2cgen.create_shared(toolchain="msvc", dirpath="./my/model")

    Later, the shared library can be referred to by its directory name:

    .. code-block:: python

        predictor = tl2cgen.Predictor(libpath="./my/model")
        # looks for ./my/model/model.dll

    Alternatively, one may specify the library down to its file name:

    .. code-block:: python

        predictor = tl2cgen.Predictor(libpath="./my/model/model.dll")
    """

    # pylint: disable=R0912

    if nthread is None or nthread <= 0:
        ncore = cpu_count()
        nthread = ncore
    dirpath = pathlib.Path(dirpath).expanduser().resolve()
    if not dirpath.exists() or not dirpath.is_dir():
        raise TL2cgenError(f"Directory {dirpath} does not exist")
    recipe = _open_and_validate_recipe(dirpath / "recipe.json")
    options = _process_options(options)

    # Write warning for potentially long compile time
    if long_build_time_warning:
        warn = False
        for source in recipe["sources"]:
            if int(source["length"]) > 10000:
                warn = True
                break
        if warn:
            warnings.warn(
                "WARNING: some of the source files are long. Expect long build time. "
                "You may want to adjust the parameter 'parallel_comp'."
            )

    tstart = time.perf_counter()
    _toolchain_exist_check(toolchain)
    if toolchain == "msvc":
        _create_shared = _create_shared_msvc
    else:
        _create_shared = _create_shared_gcc
    libpath = _create_shared(
        dirpath, toolchain, recipe, nthread=nthread, options=options, verbose=verbose
    )
    if verbose:
        elapsed_time = time.perf_counter() - tstart
        print(f"Generated shared library in {elapsed_time:.2f} seconds")
    return libpath



"""Data matrix"""
import ctypes
from typing import Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import scipy  # type: ignore

from .dtypes import (
    numpy_type_to_type_info,
    type_info_to_ctypes_type,
    type_info_to_numpy_type,
)
from .exception import TL2cgenError
from .libloader import _LIB, _check_call
from .util import c_str


class DMatrix:
    """Data matrix used in TL2cgen.

    Parameters
    ----------
    data :
        Data source
    dtype :
        If specified, the data will be casted into the corresponding data type.
    missing :
        Value in the data that represents a missing entry. If set to ``None``,
        ``numpy.nan`` will be used.
    """

    # pylint: disable=R0902,R0903,R0913

    def __init__(
        self,
        data: Union[str, npt.NDArray, scipy.sparse.csr_matrix],
        *,
        dtype: Optional[str] = None,
        missing: Optional[float] = None,
    ):
        if data is None:
            raise TL2cgenError("'data' argument cannot be None")

        self.handle = ctypes.c_void_p()

        if isinstance(data, (str,)):
            raise TL2cgenError(
                "'data' argument cannot be a string. Did you mean to load data from a text file? "
                "Please use the following packages to load the text file:\n"
                "   * CSV file: Use pandas.read_csv() or numpy.loadtxt()\n"
                "   * LIBSVM file: Use sklearn.datasets.load_svmlight_file()"
            )
        if isinstance(data, scipy.sparse.csr_matrix):
            self._init_from_csr(data, dtype=dtype)
        elif isinstance(data, scipy.sparse.csc_matrix):
            self._init_from_csr(data.tocsr(), dtype=dtype)
        elif isinstance(data, np.ndarray):
            self._init_from_npy2d(data, missing=missing, dtype=dtype)
        else:  # any type that's convertible to CSR matrix is O.K.
            try:
                csr = scipy.sparse.csr_matrix(data)
                self._init_from_csr(csr, dtype=dtype)
            except Exception as e:
                raise TypeError(
                    f"Cannot initialize DMatrix from {type(data).__name__}"
                ) from e
        num_row, num_col, nelem = self._get_dims()
        self.shape = (num_row, num_col)
        self.size = nelem

    def _init_from_csr(
        self, csr: scipy.sparse.csr_matrix, *, dtype: Optional[str] = None
    ) -> None:
        """Initialize data from a CSR (Compressed Sparse Row) matrix"""
        if len(csr.indices) != len(csr.data):
            raise ValueError(
                f"indices and data not of same length: {len(csr.indices)} vs {len(csr.data)}"
            )
        if len(csr.indptr) != csr.shape[0] + 1:
            raise ValueError(
                "len(indptr) must be equal to 1 + [number of rows]"
                f"len(indptr) = {len(csr.indptr)} vs 1 + [number of rows] = {1 + csr.shape[0]}"
            )
        if csr.indptr[-1] != len(csr.data):
            raise ValueError(
                "last entry of indptr must be equal to len(data)"
                f"indptr[-1] = {csr.indptr[-1]} vs len(data) = {len(csr.data)}"
            )

        if dtype is None:
            data_type = csr.data.dtype
        else:
            data_type = type_info_to_numpy_type(dtype)
        data_type_code = numpy_type_to_type_info(data_type)
        data_ptr_type = ctypes.POINTER(type_info_to_ctypes_type(data_type_code))
        if data_type_code not in ["float32", "float64"]:
            raise ValueError("data should be either float32 or float64 type")

        data = np.array(csr.data, copy=False, dtype=data_type, order="C")
        indices = np.array(csr.indices, copy=False, dtype=np.uintc, order="C")
        indptr = np.array(csr.indptr, copy=False, dtype=np.uintp, order="C")
        _check_call(
            _LIB.TL2cgenDMatrixCreateFromCSR(
                data.ctypes.data_as(data_ptr_type),
                c_str(data_type_code),
                indices.ctypes.data_as(ctypes.POINTER(ctypes.c_uint)),
                indptr.ctypes.data_as(ctypes.POINTER(ctypes.c_size_t)),
                ctypes.c_size_t(csr.shape[0]),
                ctypes.c_size_t(csr.shape[1]),
                ctypes.byref(self.handle),
            )
        )

    def _init_from_npy2d(
        self,
        mat: npt.NDArray,
        *,
        missing: Optional[float] = None,
        dtype: Optional[str] = None,
    ) -> None:
        """
        Initialize data from a 2-D numpy matrix.
        If ``mat`` does not have ``order='C'`` (also known as row-major) or is not
        contiguous, a temporary copy will be made.
        If ``mat`` does not have ``dtype=numpy.float32``, a temporary copy will be
        made also.
        Thus, as many as two temporary copies of data can be made. One should set
        input layout and type judiciously to conserve memory.
        """
        if len(mat.shape) != 2:
            raise ValueError("Input numpy.ndarray must be two-dimensional")
        data_type: npt.DTypeLike = (
            mat.dtype if dtype is None else type_info_to_numpy_type(dtype)
        )
        data_type_code = numpy_type_to_type_info(data_type)
        data_ptr_type = ctypes.POINTER(type_info_to_ctypes_type(data_type_code))
        if data_type_code not in ["float32", "float64"]:
            raise ValueError("data should be either float32 or float64 type")
        # flatten the array by rows and ensure it is float32.
        # we try to avoid data copies if possible
        # (reshape returns a view when possible and we explicitly tell np.array to
        #  avoid copying)
        data = np.array(mat.reshape(mat.size), copy=False, dtype=data_type)
        missing = missing if missing is not None else np.nan
        missing_ar = np.array([missing], dtype=data_type, order="C")
        _check_call(
            _LIB.TL2cgenDMatrixCreateFromMat(
                data.ctypes.data_as(data_ptr_type),
                c_str(data_type_code),
                ctypes.c_size_t(mat.shape[0]),
                ctypes.c_size_t(mat.shape[1]),
                missing_ar.ctypes.data_as(data_ptr_type),
                ctypes.byref(self.handle),
            )
        )

    def _get_dims(self) -> Tuple[int, int, int]:
        num_row = ctypes.c_size_t()
        num_col = ctypes.c_size_t()
        nelem = ctypes.c_size_t()
        _check_call(
            _LIB.TL2cgenDMatrixGetDimension(
                self.handle,
                ctypes.byref(num_row),
                ctypes.byref(num_col),
                ctypes.byref(nelem),
            )
        )
        return num_row.value, num_col.value, nelem.value

    def __del__(self):
        if self.handle:
            _check_call(_LIB.TL2cgenDMatrixFree(self.handle))
            self.handle = None

    def __repr__(self):
        return (
            f"<{self.shape[0]}x{self.shape[1]} sparse matrix of type tl2cgen.DMatrix\n"
            f"        with {self.size} stored elements in Compressed Sparse Row format>"
        )



"""Exception classes used in TL2cgen"""


class TL2cgenError(Exception):
    """Error thrown by TL2cgen"""


class TL2cgenLibraryNotFound(Exception):
    """Error thrown by when TL2cgen is not found"""



# coding: utf-8
"""
TL2cgen (TreeLite 2 C GENerator):
Model compiler for decision tree ensembles
"""
from .core import _py_version, annotate_branch, generate_c_code
from .create_shared import create_shared
from .data import DMatrix
from .exception import TL2cgenError
from .generate_makefile import generate_cmakelists, generate_makefile
from .predictor import Predictor
from .shortcuts import export_lib, export_srcpkg

__version__ = _py_version()

__all__ = [
    "annotate_branch",
    "create_shared",
    "export_lib",
    "export_srcpkg",
    "generate_c_code",
    "generate_cmakelists",
    "generate_makefile",
    "DMatrix",
    "Predictor",
    "TL2cgenError",
]



"""Core module of TL2cgen"""
import pathlib
from typing import Any, Dict, Optional, Union

import treelite

from .data import DMatrix
from .handle_class import _Annotator, _Compiler, _TreeliteModel


def _py_version() -> str:
    """Get the TL2cgen version from Python version file."""
    version_file = pathlib.Path(__file__).parent / "VERSION"
    with open(version_file, encoding="utf-8") as f:
        return f.read().strip()


def generate_c_code(
    model: treelite.Model,
    dirpath: Union[str, pathlib.Path],
    params: Optional[Dict[str, Any]],
    compiler: str = "ast_native",
    *,
    verbose: bool = False,
) -> None:
    """
    Generate prediction code from a tree ensemble model. The code will be C99
    compliant. One header file (.h) will be generated, along with one or more
    source files (.c). Use :py:meth:`create_shared` method to package
    prediction code as a dynamic shared library (.so/.dll/.dylib).

    Parameters
    ----------
    model :
        Model to convert to C code
    dirpath :
        Directory to store header and source files
    params :
        Parameters for compiler. See :py:doc:`this page </compiler_param>`
        for the list of compiler parameters.
    compiler :
        Kind of C code generator to use. Currently, there are two possible values:
        {"ast_native", "failsafe"}
    verbose :
        Whether to print extra messages during compilation

    Example
    -------
    The following populates the directory ``./model`` with source and header
    files:

    .. code-block:: python

       tl2cgen.compile(model, dirpath="./my/model", params={}, verbose=True)

    If parallel compilation is enabled (parameter ``parallel_comp``), the files
    are in the form of ``./my/model/header.h``, ``./my/model/main.c``,
    ``./my/model/tu0.c``, ``./my/model/tu1.c`` and so forth, depending on
    the value of ``parallel_comp``. Otherwise, there will be exactly two files:
    ``./model/header.h``, ``./my/model/main.c``
    """
    _model = _TreeliteModel(model)
    compiler_obj = _Compiler(params, compiler, verbose)
    compiler_obj.compile(_model, dirpath)


def annotate_branch(
    model: treelite.Model,
    dmat: DMatrix,
    path: Union[str, pathlib.Path],
    *,
    nthread: Optional[int] = None,
    verbose: bool = False,
) -> None:
    """
    Annotate branches in a given model using frequency patterns in the training data and save
    the annotation data to a JSON file. Each node gets the count of the instances that belong to it.

    Parameters
    ----------
    dmat :
        Data matrix representing the training data
    path :
        Location of JSON file
    model :
        Model to annotate
    nthread :
        Number of threads to use while annotating. If missing, use all physical cores in the system.
    verbose :
        Whether to print extra messages
    """
    _model = _TreeliteModel(model)
    nthread = nthread if nthread is not None else 0
    annotator = _Annotator(_model, dmat, nthread, verbose)
    annotator.save(path)



"""
Tools to interact with Microsoft Visual C++ (MSVC)
"""
import itertools
import os
import pathlib
import re
import sys
from typing import Any, Dict, List

from packaging.version import parse as parse_version

from .create_shared import _create_shared_base
from .util import _libext

LIBEXT = _libext()


if sys.platform != "win32":

    class WindowsError(Exception):  # pylint: disable=C0115,W0622
        pass


def _is_64bit_windows() -> bool:
    return "PROGRAMFILES(X86)" in os.environ


def _varsall_bat_path() -> pathlib.Path:  # pylint: disable=R0912
    if sys.platform != "win32":
        raise RuntimeError("_varsall_bat_path() supported only on Windows")

    # if a custom location is given, try that first
    if "TL2CGEN_VCVARSALL" in os.environ:
        candidate = pathlib.Path(os.environ["TL2CGEN_VCVARSALL"]).resolve()
        if candidate.name.lower() != "vcvarsall.bat":
            raise OSError(
                "Environment variable TL2CGEN_VCVARSALL must point to file vcvarsall.bat"
            )
        if candidate.is_file():
            return candidate
        raise OSError(
            "Environment variable TL2CGEN_VCVARSALL does not refer to existing vcvarsall.bat"
        )

    # == Bunch of heuristics to locate vcvarsall.bat ==

    # List of possible paths to vcvarsall.bat
    candidate_paths = []
    try:
        import winreg  # pylint: disable=E0401,C0415

        if _is_64bit_windows():
            key_name = r"SOFTWARE\Wow6432Node\Microsoft\VisualStudio\SxS\VS7"
        else:
            key_name = r"SOFTWARE\Microsoft\VisualStudio\SxS\VC7"
        key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, key_name)
        i = 0
        while True:
            try:
                version, vcroot, _ = winreg.EnumValue(key, i)
                vcroot = pathlib.Path(vcroot).resolve()
                if parse_version(version) >= parse_version("15.0"):
                    # Visual Studio 2017 revamped directory structure
                    candidate_paths.append(vcroot / r"VC\Auxiliary\Build\vcvarsall.bat")
                else:
                    candidate_paths.append(vcroot / r"VC\vcvarsall.bat")
            except WindowsError:  # pylint: disable=E0602
                break
            i += 1
    except FileNotFoundError:
        pass  # No registry key found
    except ImportError:
        pass  # No winreg module

    for candidate in candidate_paths:
        if candidate.is_file():
            return candidate

    # If registry method fails, try a bunch of pre-defined paths

    # Visual Studio 2017 and higher
    candidate_paths = list(
        itertools.chain(
            pathlib.Path(r"C:\Program Files (x86)\Microsoft Visual Studio").glob("*"),
            pathlib.Path(r"C:\Program Files\Microsoft Visual Studio").glob("*"),
        )
    )
    for vcroot in candidate_paths:
        if re.fullmatch(r"[0-9]+", vcroot.name):
            for candidate in vcroot.glob(r"*\VC\Auxiliary\Build\vcvarsall.bat"):
                if candidate.is_file():
                    return candidate
    # Previous versions of Visual Studio
    candidate_paths = list(
        itertools.chain(
            pathlib.Path(r"C:\Program Files (x86)").glob(
                r"Microsoft Visual Studio*\VC\vcvarsall.bat"
            ),
            pathlib.Path(r"C:\Program Files").glob(
                r"Microsoft Visual Studio*\VC\vcvarsall.bat"
            ),
        )
    )
    for candidate in candidate_paths:
        if candidate.is_file():
            return candidate

    raise OSError(
        "vcvarsall.bat not found; please specify its full path in the environment "
        "variable TL2CGEN_VCVARSALL"
    )


def _obj_ext() -> str:
    return ".obj"


# pylint: disable=W0613
def _obj_cmd(
    source: str,
    toolchain: str,
    options: List[str],
):
    source_file = source + ".c"
    options_str = " ".join(options)
    return f"cl.exe /c /openmp /Ox {source_file} {options_str}"


# pylint: disable=W0613
def _lib_cmd(
    objects: List[str],
    target: str,
    lib_ext: str,
    toolchain: str,
    options: List[str],
) -> str:
    objects_str = " ".join(objects)
    options_str = " ".join(options)
    return f"cl.exe /LD /Fe{target} /openmp {objects_str} {options_str}"


# pylint: disable=R0913
def _create_shared_msvc(
    dirpath: pathlib.Path,
    toolchain: str,
    recipe: Dict[str, Any],
    *,
    nthread: int,
    options: List[str],
    verbose: bool,
) -> pathlib.Path:
    # Specify command to compile an object file
    recipe["object_ext"] = _obj_ext()
    recipe["library_ext"] = LIBEXT

    # pylint: disable=R0801

    def _obj_cmd_wrapped(source: str) -> str:
        return _obj_cmd(source, toolchain, options)

    def _lib_cmd_wrapped(objects: List[str], target: str) -> str:
        return _lib_cmd(objects, target, LIBEXT, toolchain, options)

    recipe["create_object_cmd"] = _obj_cmd_wrapped
    recipe["create_library_cmd"] = _lib_cmd_wrapped
    plat_target = "amd64" if _is_64bit_windows() else "x86"
    recipe["initial_cmd"] = f'"{_varsall_bat_path()}" {plat_target}'
    return _create_shared_base(dirpath, recipe, nthread=nthread, verbose=verbose)



"""Utility functions"""
import ctypes
import json
import pathlib
from typing import Any, Dict, List, Optional

from .exception import TL2cgenError


def c_str(string):
    """Convert a Python string to C string"""
    return ctypes.c_char_p(string.encode("utf-8"))


def py_str(string):
    """Convert C string back to Python string"""
    return string.decode("utf-8")


def _open_and_validate_recipe(recipe_path: pathlib.Path) -> Dict[str, Any]:
    """Ensure that the build recipe contains necessary fields"""
    try:
        with open(recipe_path, "r", encoding="utf-8") as f:
            recipe = json.load(f)
    except IOError as e:
        raise TL2cgenError("Failed to open recipe.json") from e
    if "sources" not in recipe or "target" not in recipe:
        raise TL2cgenError("Malformed recipe.json")
    return recipe


def _process_options(options: Optional[List[str]]) -> List[str]:
    """Ensure that options are in the form of list of strings."""
    if options is not None:
        try:
            _ = iter(options)
            options = [str(x) for x in options]
        except TypeError as e:
            raise TL2cgenError("options must be a list of string") from e
    else:
        options = []
    return options



"""Utility functions to handle types"""

import ctypes
from typing import Any, Dict

import numpy as np
import numpy.typing as npt

_CTYPES_TYPE_TABLE: Dict[str, Any] = {
    "uint32": ctypes.c_uint32,
    "float32": ctypes.c_float,
    "float64": ctypes.c_double,
}

_NUMPY_TYPE_TABLE: Dict[str, npt.DTypeLike] = {
    "uint32": np.uint32,
    "float32": np.float32,
    "float64": np.float64,
}


def type_info_to_ctypes_type(type_info: str) -> Any:
    """Obtain ctypes type corresponding to a given TypeInfo"""
    return _CTYPES_TYPE_TABLE[type_info]


def type_info_to_numpy_type(type_info: str) -> npt.DTypeLike:
    """Obtain ctypes type corresponding to a given TypeInfo"""
    return _NUMPY_TYPE_TABLE[type_info]


def numpy_type_to_type_info(type_info: npt.DTypeLike) -> str:
    """Obtain TypeInfo corresponding to a given NumPy type"""
    if type_info == np.uint32:
        return "uint32"
    if type_info == np.float32:
        return "float32"
    if type_info == np.float64:
        return "float64"
    raise ValueError(f"Unrecognized NumPy type: {type_info}")



"""Utilities for contrib module"""
import os
import subprocess
from sys import platform as _platform


def _is_windows() -> bool:
    return _platform == "win32"


def _shell() -> str:
    if _is_windows():
        return "cmd.exe"
    if "SHELL" in os.environ:
        return os.environ["SHELL"]
    return "/bin/sh"  # use POSIX-compliant shell if SHELL is not set


def _libext():
    if _platform == "darwin":
        return ".dylib"
    if _platform in ("win32", "cygwin"):
        return ".dll"
    return ".so"


def _toolchain_exist_check(toolchain: str) -> None:
    if toolchain != "msvc":
        retcode = subprocess.call(
            f"{toolchain} --version",
            shell=True,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if retcode != 0:
            raise ValueError(
                f"Toolchain {toolchain} not found. Ensure that it is installed and "
                "that it is a variant of GCC or Clang."
            )


def _create_log_cmd_unix(logfile: str) -> str:
    return f"true > {logfile}"


def _save_retcode_cmd_unix(logfile: str) -> str:
    if _shell().endswith("fish"):  # special handling for fish shell
        return f"echo $status >> {logfile}"
    return f"echo $? >> {logfile}"


def _create_log_cmd_windows(logfile: str) -> str:
    return f"type NUL > {logfile}"


def _save_retcode_cmd_windows(logfile: str) -> str:
    return f"echo %errorlevel% >> {logfile}"



"""Internal classes to hold native handles"""
import ctypes
import json
import pathlib
from typing import Any, Dict, Optional, Union

import numpy as np
import treelite

from .data import DMatrix
from .dtypes import type_info_to_ctypes_type
from .libloader import _LIB, _check_call
from .util import c_str, py_str


class _TreeliteModel:
    """
    Internal class holding a handle to Treelite model. We maintain a separate internal class,
    to maintain **loose coupling** between Treelite and TL2cgen. This way, TL2cgen can support
    past and future versions of Treelite (within the same major version).
    """

    def __init__(self, model: treelite.Model):
        model_bytes = model.serialize_bytes()
        model_bytes_len = len(model_bytes)
        buffer = ctypes.create_string_buffer(model_bytes, model_bytes_len)
        self.handle = ctypes.c_void_p()
        _check_call(
            _LIB.TL2cgenLoadTreeliteModelFromBytes(
                ctypes.pointer(buffer),
                ctypes.c_size_t(model_bytes_len),
                ctypes.byref(self.handle),
            )
        )
        major_ver, minor_ver, patch_ver = (
            ctypes.c_int32(),
            ctypes.c_int32(),
            ctypes.c_int32(),
        )
        _check_call(
            _LIB.TL2cgenQueryTreeliteModelVersion(
                self.handle,
                ctypes.byref(major_ver),
                ctypes.byref(minor_ver),
                ctypes.byref(patch_ver),
            )
        )
        self.__version__ = f"{major_ver.value}.{minor_ver.value}.{patch_ver.value}"

    def __del__(self):
        if self.handle:
            _check_call(_LIB.TL2cgenFreeTreeliteModel(self.handle))
            self.handle = None


class _Annotator:
    """Annotator object"""

    def __init__(
        self,
        model: _TreeliteModel,
        dmat: DMatrix,
        nthread: int,
        verbose: bool = False,
    ):
        self.handle = ctypes.c_void_p()
        _check_call(
            _LIB.TL2cgenAnnotateBranch(
                model.handle,
                dmat.handle,
                ctypes.c_int(nthread),
                ctypes.c_int(1 if verbose else 0),
                ctypes.byref(self.handle),
            )
        )

    def save(self, path: Union[str, pathlib.Path]):
        """Save annotation data to a JSON file"""
        path = pathlib.Path(path).expanduser().resolve()
        _check_call(_LIB.TL2cgenAnnotationSave(self.handle, c_str(str(path))))

    def __del__(self):
        if self.handle:
            _check_call(_LIB.TL2cgenAnnotationFree(self.handle))
            self.handle = None


class _Compiler:
    """Compiler object"""

    def __init__(
        self,
        params: Optional[Dict[str, Any]],
        compiler: str = "ast_native",
        verbose: bool = False,
    ):
        self.handle = ctypes.c_void_p()
        if params is None:
            params = {}
        if verbose:
            params["verbose"] = 1
        if isinstance(params.get("annotate_in"), pathlib.Path):
            params["annotate_in"] = str(params["annotate_in"])
        params_json_str = json.dumps(params)
        _check_call(
            _LIB.TL2cgenCompilerCreate(
                c_str(compiler), c_str(params_json_str), ctypes.byref(self.handle)
            )
        )

    def compile(self, model: _TreeliteModel, dirpath: Union[str, pathlib.Path]) -> None:
        """Generate prediction code"""
        dirpath = pathlib.Path(dirpath).expanduser().resolve()
        _check_call(
            _LIB.TL2cgenCompilerGenerateCode(
                self.handle, model.handle, c_str(str(dirpath))
            )
        )

    def __del__(self):
        if self.handle:
            _check_call(_LIB.TL2cgenCompilerFree(self.handle))
            self.handle = None


class _OutputVector:
    """Output vector object, used to hold prediction results from Predictor"""

    def __init__(
        self,
        predictor_handle: ctypes.c_void_p,
        dmat_handle: ctypes.c_void_p,
    ):
        self.handle = ctypes.c_void_p()
        _check_call(
            _LIB.TL2cgenPredictorCreateOutputVector(
                predictor_handle, dmat_handle, ctypes.byref(self.handle)
            )
        )
        type_str = ctypes.c_char_p()
        _check_call(
            _LIB.TL2cgenPredictorQueryLeafOutputType(
                predictor_handle, ctypes.byref(type_str)
            )
        )
        self.typestr_ = py_str(type_str.value)
        length = ctypes.c_size_t()
        _check_call(
            _LIB.TL2cgenPredictorQueryResultSize(
                predictor_handle, dmat_handle, ctypes.byref(length)
            )
        )
        self.length_ = length.value

    def __del__(self):
        if self.handle:
            _check_call(_LIB.TL2cgenPredictorDeleteOutputVector(self.handle))
            self.handle = None

    def toarray(self):
        """Convert to NumPy array"""
        ptr = ctypes.c_void_p()
        _check_call(
            _LIB.TL2cgenPredictorGetRawPointerFromOutputVector(
                self.handle, ctypes.byref(ptr)
            )
        )
        ptr_type = ctypes.POINTER(type_info_to_ctypes_type(self.typestr_))
        casted_ptr = ctypes.cast(ptr, ptr_type)
        return np.copy(
            np.ctypeslib.as_array(casted_ptr, shape=(self.length_,)), order="C"
        )



"""Logic for launching C compiler to build shared libs"""
import pathlib
import subprocess
from typing import Any, Dict

from ..exception import TL2cgenError
from .util import (
    _create_log_cmd_unix,
    _create_log_cmd_windows,
    _is_windows,
    _save_retcode_cmd_unix,
    _save_retcode_cmd_windows,
    _shell,
)


def _enqueue(args: Dict[str, Any]) -> subprocess.Popen:
    tid = args["tid"]
    queue = args["queue"]
    dirpath = args["dirpath"]
    init_cmd = args["init_cmd"]
    create_log_cmd = args["create_log_cmd"]
    save_retcode_cmd = args["save_retcode_cmd"]

    # pylint: disable=R1732
    proc = subprocess.Popen(
        _shell(),
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=dirpath,
    )
    assert proc.stdin is not None
    proc.stdin.write((init_cmd + "\n").encode())
    proc.stdin.write(create_log_cmd(f"retcode_cpu{tid}.txt" + "\n").encode())
    for command in queue:
        proc.stdin.write((command + "\n").encode())
        proc.stdin.write((save_retcode_cmd(f"retcode_cpu{tid}.txt") + "\n").encode())
    proc.stdin.flush()

    return proc


def _wait(proc: subprocess.Popen, args: Dict[str, Any]):
    tid = args["tid"]
    dirpath = args["dirpath"]
    stdout, _ = proc.communicate()
    with open(dirpath / f"retcode_cpu{tid}.txt", "r", encoding="UTF-8") as f:
        retcode = [int(line) for line in f]
    return {"stdout": stdout.decode(), "retcode": retcode}


def _create_shared_base(
    dirpath: pathlib.Path,
    recipe: Dict[str, Any],
    *,
    nthread: int,
    verbose: bool,
):  # pylint: disable=R0914
    # Fetch toolchain-specific commands
    obj_cmd = recipe["create_object_cmd"]
    lib_cmd = recipe["create_library_cmd"]
    create_log_cmd = _create_log_cmd_windows if _is_windows() else _create_log_cmd_unix
    save_retcode_cmd = (
        _save_retcode_cmd_windows if _is_windows() else _save_retcode_cmd_unix
    )

    # 1. Compile sources in parallel
    if verbose:
        obj_ext = recipe["object_ext"]
        print(
            f"Compiling sources files in directory {dirpath} into object files (*{obj_ext})..."
        )
    workqueue = [
        {
            "tid": tid,
            "queue": [],
            "dirpath": dirpath,
            "init_cmd": recipe["initial_cmd"],
            "create_log_cmd": create_log_cmd,
            "save_retcode_cmd": save_retcode_cmd,
        }
        for tid in range(nthread)
    ]
    for i, source in enumerate(recipe["sources"]):
        workqueue[i % nthread]["queue"].append(obj_cmd(source["name"]))
    proc = [_enqueue(workqueue[tid]) for tid in range(nthread)]
    result = []
    for tid in range(nthread):
        result.append(_wait(proc[tid], workqueue[tid]))

    for tid in range(nthread):
        if not all(x == 0 for x in result[tid]["retcode"]):
            log_path = dirpath / f"log_cpu{tid}.txt"
            with open(log_path, "w", encoding="UTF-8") as f:
                f.write(result[tid]["stdout"] + "\n")
            raise TL2cgenError(
                f"Error occured in worker #{tid}: " + result[tid]["stdout"]
            )

    # 2. Package objects into a dynamic shared library
    full_libpath = dirpath.joinpath(recipe["target"] + recipe["library_ext"])
    if verbose:
        print(f"Generating dynamic shared library {full_libpath}...")
    objects = [x["name"] + recipe["object_ext"] for x in recipe["sources"]]
    objects.extend(recipe.get("extra", []))
    workqueue = [
        {
            "tid": 0,
            "queue": [lib_cmd(objects, recipe["target"])],
            "dirpath": dirpath,
            "init_cmd": recipe["initial_cmd"],
            "create_log_cmd": create_log_cmd,
            "save_retcode_cmd": save_retcode_cmd,
        }
    ]
    proc = [_enqueue(workqueue[0])]
    result = [_wait(proc[0], workqueue[0])]

    if result[0]["retcode"][0] != 0:
        with open(dirpath / "log_cpu0.txt", "w", encoding="UTF-8") as f:
            f.write(result[0]["stdout"] + "\n")
        raise TL2cgenError(
            "Error occured while creating dynamic library: " + result[0]["stdout"]
        )

    # 3. Clean up
    for tid in range(nthread):
        dirpath.joinpath(f"retcode_cpu{tid}.txt").unlink()

    # Return full path of shared library
    return dirpath.joinpath(full_libpath)



"""
Predictor module
"""
import ctypes
import pathlib
from typing import Optional, Union

from .contrib.util import _libext
from .data import DMatrix
from .exception import TL2cgenError
from .handle_class import _OutputVector
from .libloader import _LIB, _check_call
from .util import c_str, py_str


class Predictor:
    """
    Predictor class is a convenient wrapper for loading shared libs.
    TL2cgen uses OpenMP to launch multiple CPU threads to perform predictions
    in parallel.

    Parameters
    ----------
    libpath :
        location of dynamic shared library (.dll/.so/.dylib)
    nthread :
        number of worker threads to use; if unspecified, use maximum number of
        hardware threads
    verbose :
        Whether to print extra messages during construction
    """

    def __init__(
        self,
        libpath: Union[str, pathlib.Path],
        *,
        nthread: Optional[int] = None,
        verbose: bool = False,
    ):
        nthread = nthread if nthread is not None else -1
        libpath = pathlib.Path(libpath).expanduser().resolve()
        if libpath.is_dir():
            # directory is given; locate shared library inside it
            resolved_libpath = None
            ext = _libext()
            for candidate in libpath.glob(f"*{ext}"):
                try:
                    resolved_libpath = candidate.resolve(strict=True)
                    break
                except FileNotFoundError:
                    continue
            if not resolved_libpath:
                raise TL2cgenError(
                    f"Directory {libpath} doesn't appear to have any dynamic shared library."
                )
        else:  # libpath is actually the name of shared library file
            if not libpath.exists():
                raise TL2cgenError(f"Shared library not found at location {libpath}")
            resolved_libpath = libpath

        self.handle = ctypes.c_void_p()
        _check_call(
            _LIB.TL2cgenPredictorLoad(
                c_str(str(resolved_libpath)),
                ctypes.c_int(nthread),
                ctypes.byref(self.handle),
            )
        )
        self._load_metadata(self.handle)

        if verbose:
            print(
                f"Dynamic shared library {resolved_libpath} has been successfully "
                "loaded into memory"
            )

    def __del__(self):
        if self.handle:
            _check_call(_LIB.TL2cgenPredictorFree(self.handle))
            self.handle = None

    @property
    def num_feature(self):
        """Query number of features used in the model"""
        return self.num_feature_

    @property
    def num_class(self):
        """Query number of output groups of the model"""
        return self.num_class_

    @property
    def pred_transform(self):
        """Query pred transform of the model"""
        return self.pred_transform_

    @property
    def global_bias(self):
        """Query global bias of the model"""
        return self.global_bias_

    @property
    def sigmoid_alpha(self):
        """Query sigmoid alpha of the model"""
        return self.sigmoid_alpha_

    @property
    def ratio_c(self):
        """Query sigmoid alpha of the model"""
        return self.ratio_c_

    @property
    def threshold_type(self):
        """Query threshold type of the model"""
        return self.threshold_type_

    @property
    def leaf_output_type(self):
        """Query threshold type of the model"""
        return self.leaf_output_type_

    def predict(
        self,
        dmat: DMatrix,
        *,
        verbose: bool = False,
        pred_margin: bool = False,
    ):
        """
        Perform batch prediction with a 2D sparse data matrix. Worker threads will
        internally divide up work for batch prediction. **Note that this function
        may be called by only one thread at a time.**

        Parameters
        ----------
        dmat:
            Batch of rows for which predictions will be made
        verbose :
            Whether to print extra messages during prediction
        pred_margin:
            Whether to produce raw margins rather than transformed probabilities
        """
        if not isinstance(dmat, DMatrix):
            raise TL2cgenError("dmat must be of type DMatrix")
        result_size = ctypes.c_size_t()
        _check_call(
            _LIB.TL2cgenPredictorQueryResultSize(
                self.handle, dmat.handle, ctypes.byref(result_size)
            )
        )

        out_result = _OutputVector(
            predictor_handle=self.handle,
            dmat_handle=dmat.handle,
        )
        out_result_size = ctypes.c_size_t()
        _check_call(
            _LIB.TL2cgenPredictorPredictBatch(
                self.handle,
                dmat.handle,
                ctypes.c_int(1 if verbose else 0),
                ctypes.c_int(1 if pred_margin else 0),
                out_result.handle,
                ctypes.byref(out_result_size),
            )
        )

        out_result_array = out_result.toarray()
        idx = int(out_result_size.value)
        res = out_result_array[0:idx].reshape((dmat.shape[0], -1))
        if self.num_class_ > 1 and dmat.shape[0] != idx:
            res = res.reshape((-1, self.num_class_))

        return res

    def _load_metadata(self, handle: ctypes.c_void_p) -> None:
        # Save # of features
        num_feature = ctypes.c_size_t()
        _check_call(
            _LIB.TL2cgenPredictorQueryNumFeature(handle, ctypes.byref(num_feature))
        )
        self.num_feature_ = num_feature.value
        # Save # of classes
        num_class = ctypes.c_size_t()
        _check_call(_LIB.TL2cgenPredictorQueryNumClass(handle, ctypes.byref(num_class)))
        self.num_class_ = num_class.value
        # Save # of pred transform
        pred_transform = ctypes.c_char_p()
        _check_call(
            _LIB.TL2cgenPredictorQueryPredTransform(
                handle, ctypes.byref(pred_transform)
            )
        )
        self.pred_transform_ = py_str(pred_transform.value)
        # Save # of sigmoid alpha
        sigmoid_alpha = ctypes.c_float()
        _check_call(
            _LIB.TL2cgenPredictorQuerySigmoidAlpha(handle, ctypes.byref(sigmoid_alpha))
        )
        self.sigmoid_alpha_ = sigmoid_alpha.value
        # Save # of ratio C
        ratio_c = ctypes.c_float()
        _check_call(_LIB.TL2cgenPredictorQueryRatioC(handle, ctypes.byref(ratio_c)))
        self.ratio_c_ = ratio_c.value
        # Save # of global bias
        global_bias = ctypes.c_float()
        _check_call(
            _LIB.TL2cgenPredictorQueryGlobalBias(handle, ctypes.byref(global_bias))
        )
        self.global_bias_ = global_bias.value
        # Save threshold type
        threshold_type = ctypes.c_char_p()
        _check_call(
            _LIB.TL2cgenPredictorQueryThresholdType(
                handle, ctypes.byref(threshold_type)
            )
        )
        self.threshold_type_ = py_str(threshold_type.value)
        # Save leaf output type
        leaf_output_type = ctypes.c_char_p()
        _check_call(
            _LIB.TL2cgenPredictorQueryLeafOutputType(
                handle, ctypes.byref(leaf_output_type)
            )
        )
        self.leaf_output_type_ = py_str(leaf_output_type.value)



"""
Tools to interact with toolchains GCC, Clang, and other UNIX compilers
"""
import pathlib
from typing import Any, Dict, List

from .create_shared import _create_shared_base
from .util import _libext

LIBEXT = _libext()


def _obj_ext() -> str:
    return ".o"


def _obj_cmd(
    source: str,
    toolchain: str,
    options: List[str],
) -> str:
    obj_ext = _obj_ext()
    source_file = source + ".c"
    obj_file = source + obj_ext
    options_str = " ".join(options)
    return (
        f"{toolchain} -c -O3 -o {obj_file} {source_file} -fPIC -std=c99 {options_str}"
    )


def _lib_cmd(
    objects: List[str],
    target: str,
    lib_ext: str,
    toolchain: str,
    options: List[str],
) -> str:
    objects_str = " ".join(objects)
    options_str = " ".join(options)
    return f"{toolchain} -shared -O3 -o {target + lib_ext} {objects_str} -std=c99 {options_str}"


def _create_shared_gcc(
    dirpath: pathlib.Path,
    toolchain: str,
    recipe: Dict[str, Any],
    *,
    nthread: int,
    options: List[str],
    verbose: bool,
) -> pathlib.Path:
    options += ["-lm"]
    # Specify command to compile an object file
    recipe["object_ext"] = _obj_ext()
    recipe["library_ext"] = LIBEXT

    # pylint: disable=R0801

    def _obj_cmd_wrapped(source: str) -> str:
        return _obj_cmd(source, toolchain, options)

    def _lib_cmd_wrapped(objects: List[str], target: str) -> str:
        return _lib_cmd(objects, target, LIBEXT, toolchain, options)

    recipe["create_object_cmd"] = _obj_cmd_wrapped
    recipe["create_library_cmd"] = _lib_cmd_wrapped
    recipe["initial_cmd"] = ""
    return _create_shared_base(dirpath, recipe, nthread=nthread, verbose=verbose)



"""Convenience functions"""
import pathlib
import shutil
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional, Union

import treelite

from .contrib.util import _toolchain_exist_check
from .core import generate_c_code
from .create_shared import create_shared
from .generate_makefile import generate_cmakelists, generate_makefile


def export_lib(
    model: treelite.Model,
    toolchain: str,
    libpath: Union[str, pathlib.Path],
    params: Optional[Dict[str, Any]] = None,
    compiler: str = "ast_native",
    *,
    nthread: Optional[int] = None,
    verbose: bool = False,
    options: Optional[List[str]] = None,
):
    """
    Convenience function: Generate prediction code and immediately turn it
    into a dynamic shared library. A temporary directory will be created to
    hold the source files.

    Parameters
    ----------
    model :
        Model to convert to C code
    toolchain :
        Which toolchain to use. You may choose one of 'msvc', 'clang', and 'gcc'.
        You may also specify a specific variation of clang or gcc (e.g. 'gcc-7')
    libpath :
        Location to save the generated dynamic shared library
    params :
        Parameters to be passed to the compiler. See
        :py:doc:`this page </compiler_param>` for the list of compiler
        parameters.
    compiler :
        Kind of C code generator to use. Currently, there are two possible values:
        {"ast_native", "failsafe"}
    nthread :
        Number of threads to use in creating the shared library.
        Defaults to the number of cores in the system.
    verbose :
        Whether to produce extra messages
    options :
        Additional options to pass to toolchain

    Example
    -------
    The one-line command

    .. code-block:: python

        tl2cgen.export_lib(model, toolchain="msvc", libpath="./mymodel.dll",
                           params={})

    is equivalent to the following sequence of commands:

    .. code-block:: python

        tl2cgen.generate_c_code(model, dirpath="/temporary/directory",
                                params={})
        tl2cgen.create_shared(toolchain="msvc",
                              dirpath="/temporary/directory")
        # Move the library out of the temporary directory
        shutil.move("/temporary/directory/mymodel.dll", "./mymodel.dll")

    """
    _toolchain_exist_check(toolchain)
    libpath = pathlib.Path(libpath).expanduser().resolve()
    long_build_time_warning = not (params and "parallel_comp" in params)

    with TemporaryDirectory() as tempdir:
        generate_c_code(model, tempdir, params, compiler, verbose=verbose)
        temp_libpath = create_shared(
            toolchain,
            tempdir,
            nthread=nthread,
            verbose=verbose,
            options=options,
            long_build_time_warning=long_build_time_warning,
        )
        if libpath.is_file():
            libpath.unlink()
        shutil.move(temp_libpath, libpath)


def export_srcpkg(
    model: treelite.Model,
    toolchain: str,
    pkgpath: Union[str, pathlib.Path],
    libname: str,
    params: Optional[Dict[str, Any]] = None,
    compiler: str = "ast_native",
    *,
    verbose: bool = False,
    options: Optional[List[str]] = None,
):  # pylint: disable=R0913
    """
    Convenience function: Generate prediction code and create a zipped source
    package for deployment. The resulting zip file will also contain a Makefile
    (or CMakeLists.txt, if you set toolchain="cmake").

    Parameters
    ----------
    model :
        Model to convert to C code
    toolchain :
        Which toolchain to use. You may choose one of "msvc", "clang", "gcc", and "cmake".
        You may also specify a specific variation of clang or gcc (e.g. "gcc-7")
    pkgpath :
        Location to save the zipped source package
    libname :
        Name of model shared library to be built
    params :
        Parameters to be passed to the compiler. See
        :py:doc:`this page </compiler_param>` for the list of compiler
        parameters.
    compiler :
        Name of compiler to use in C code generation
    verbose :
        Whether to produce extra messages
    options :
        Additional options to pass to toolchain

    Example
    -------
    The one-line command

    .. code-block:: python

        tl2cgen.export_srcpkg(model, toolchain="gcc",
                              pkgpath="./mymodel_pkg.zip",
                              libname="mymodel.so", params={})

    is equivalent to the following sequence of commands:

    .. code-block:: python

        tl2cgen.generate_c_code(model, dirpath="/temporary/directory/mymodel",
                                params={})
        tl2cgen.generate_makefile(dirpath="/temporary/directory/mymodel",
                                  toolchain="gcc")
        # Zip the directory containing C code and Makefile
        shutil.make_archive(base_name="./mymodel_pkg", format="zip",
                            root_dir="/temporary/directory",
                            base_dir="mymodel/")
    """
    # Check for file extension
    pkgpath = pathlib.Path(pkgpath).expanduser().resolve()
    if pkgpath.suffix != ".zip":
        raise ValueError("Source package file should have .zip extension")
    if toolchain != "cmake":
        _toolchain_exist_check(toolchain)

    with TemporaryDirectory() as temp_dir:
        target = pathlib.Path(libname).stem
        # Create a child directory to get desired name for target
        dirpath = pathlib.Path(temp_dir) / target
        dirpath.mkdir()
        if params is None:
            params = {}
        params["native_lib_name"] = target
        generate_c_code(model, dirpath, params, compiler, verbose=verbose)
        if toolchain == "cmake":
            generate_cmakelists(dirpath, options)
        else:
            generate_makefile(dirpath, toolchain, options)
        shutil.make_archive(
            base_name=str(pkgpath.with_suffix("")),
            format="zip",
            root_dir=temp_dir,
            base_dir=target,
        )



# coding: utf-8
"""Find the path to TL2cgen dynamic library files."""

import ctypes
import os
import pathlib
import platform
import sys
import warnings
from typing import List

from .exception import TL2cgenError, TL2cgenLibraryNotFound
from .util import py_str


def _find_lib_path() -> List[pathlib.Path]:
    """Find the path to TL2cgen dynamic library files.

    Returns
    -------
    lib_path
       List of all found library path to TL2cgen
    """
    curr_path = pathlib.Path(__file__).expanduser().absolute().parent
    dll_path = [
        # When installed, libtreelite will be installed in <site-package-dir>/lib
        curr_path / "lib",
        # Editable installation
        curr_path.parent.parent / "build",
        # Use libtreelite from a system prefix, if available. This should be the last option.
        pathlib.Path(sys.prefix).absolute().resolve() / "lib",
    ]

    if sys.platform == "win32":
        if platform.architecture()[0] == "64bit":
            dll_path.append(curr_path.joinpath("../../windows/x64/Release/"))
            dll_path.append(curr_path.joinpath("./windows/x64/Release/"))
        else:
            dll_path.append(curr_path.joinpath("../../windows/Release/"))
            dll_path.append(curr_path.joinpath("./windows/Release/"))
        dll_path = [p.joinpath("tl2cgen.dll") for p in dll_path]
    elif sys.platform.startswith(("linux", "freebsd", "emscripten", "OS400")):
        dll_path = [p.joinpath("libtl2cgen.so") for p in dll_path]
    elif sys.platform == "darwin":
        dll_path = [p.joinpath("libtl2cgen.dylib") for p in dll_path]
    elif sys.platform == "cygwin":
        dll_path = [p.joinpath("cygtl2cgen.dll") for p in dll_path]
    else:
        raise RuntimeError(f"Unrecognized platform: {sys.platform}")

    lib_path = [p for p in dll_path if p.exists() and p.is_file()]

    # TL2CGEN_BUILD_DOC is defined by sphinx conf.
    if not lib_path and not os.environ.get("TL2CGEN_BUILD_DOC", False):
        link = "https://tl2cgen.readthedocs.io/en/latest/install.html"
        msg = (
            "Cannot find TL2cgen Library in the candidate path.  "
            + "List of candidates:\n- "
            + ("\n- ".join(str(x) for x in dll_path))
            + "\nTL2cgen Python package path: "
            + str(curr_path)
            + "\nsys.prefix: "
            + sys.prefix
            + "\nSee: "
            + link
            + " for installing TL2cgen."
        )
        raise TL2cgenLibraryNotFound(msg)
    return lib_path


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _log_callback(msg: bytes) -> None:
    """Redirect logs from native library into Python console"""
    print(msg.decode("utf-8"))


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _warn_callback(msg: bytes) -> None:
    """Redirect warnings from native library into Python console"""
    warnings.warn(msg.decode("utf-8"))


def _load_lib() -> ctypes.CDLL:
    """Load TL2cgen Library."""
    lib_paths = _find_lib_path()
    if not lib_paths:
        # This happens only when building document.
        return None  # type: ignore
    if sys.version_info >= (3, 8) and sys.platform == "win32":
        # pylint: disable=no-member
        os.add_dll_directory(
            os.path.join(os.path.normpath(sys.prefix), "Library", "bin")
        )
    os_error_list = []
    lib = None
    for lib_path in lib_paths:
        try:
            lib = ctypes.cdll.LoadLibrary(str(lib_path))
            setattr(lib, "path", lib_path)
        except OSError as e:
            os_error_list.append(str(e))
            continue
    if not lib:
        libname = lib_paths[0].name
        raise TL2cgenError(
            f"""
TL2cgen Library ({libname}) could not be loaded.
Likely causes:
  * OpenMP runtime is not installed
    - vcomp140.dll or libgomp-1.dll for Windows
    - libomp.dylib for Mac OSX
    - libgomp.so for Linux and other UNIX-like OSes
    Mac OSX users: Run `brew install libomp` to install OpenMP runtime.
  * You are running 32-bit Python on a 64-bit OS
Error message(s): {os_error_list}
"""
        )
    lib.TL2cgenGetLastError.restype = ctypes.c_char_p
    lib.log_callback = _log_callback  # type: ignore
    lib.warn_callback = _warn_callback  # type: ignore
    if lib.TL2cgenRegisterLogCallback(lib.log_callback) != 0:
        raise TL2cgenError(py_str(lib.TL2cgenGetLastError()))
    if lib.TL2cgenRegisterWarningCallback(lib.warn_callback) != 0:
        raise TL2cgenError(py_str(lib.TL2cgenGetLastError()))
    return lib


def _check_call(ret):
    """Check the return value of C API call

    This function will raise exception when error occurs.
    Wrap every API call with this function

    Parameters
    ----------
    ret : int
        return value from API calls
    """
    if ret != 0:
        raise TL2cgenError(_LIB.TL2cgenGetLastError().decode("utf-8"))


# Load native library in the global scope
_LIB = _load_lib()
